
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <initng.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include <initng_global.h>
#include <initng_active_state.h>
#include <initng_active_db.h>
#include <initng_process_db.h>
#include <initng_service_cache.h>
#include <initng_handler.h>
#include <main.h>
#include <initng_active_db.h>
#include <initng_toolbox.h>
#include <initng_plugin_hook.h>
#include <initng_load_module.h>
#include <initng_plugin_callers.h>
#include <initng_error.h>
#include <initng_plugin.h>
#include <initng_static_states.h>
#include <initng_control_command.h>

#include <initng-paths.h>

#include "initng_ngc4.h"

#undef D_
#define D_ W_

static void accepted_client(f_module_h * from);
static void closesock(void);
static void handle_client(int fd);
static int sendping(void);
static int open_socket(void);
static void check_socket(void);


/*
   In the Linux implementation, sockets which are visible in the file system
   honour the permissions of the directory they are in.  Their owner, group
   and their permissions can be changed.  Creation of a new socket will
   fail if the process does not have write and search (execute) permission
   on the directory the socket is created in.  Connecting to the socket
   object requires read/write permission.  This behavior differs from  many
   BSD derived systems which ignore permissions for Unix sockets.  Portable
   programs should not rely on this feature for security.
 */


/* globals */
struct stat sock_stat;
const char *socket_filename;

f_module_h fdh = { &accepted_client, -1 };


static void closesock(void)
{
    /* Check if we need to remove hooks */
    if (fdh.fds < 0)
        return;
    D_("closesock %d\n", fdh.fds);

    /* close socket and set to 0 */
    close(fdh.fds);
    fdh.fds = -1;
}


/* this is called on protocol_missmatch to reload initng if that plugin is present */
static void initng_reload(void)
{
        s_command *reload;


        /*
         * Now we are going to try reload initng
         * 
         * ngc -c  is the reload command
         */
        reload = initng_command_find_by_command_id('c');
        if (reload && reload->u.void_command_call)
        {
            /* call the reload call in that s_command */
            (*reload->u.void_command_call) (NULL);
        }
}

static void handle_client(int fd)
{
    result_desc * result = NULL;
    read_header header;
    void *header_data = NULL;
    ssize_t sent=0;

    s_command *tmp_cmd;         /* temporary storage for a command */

    assert(fd>0);

    D_("handle_client(%i);\n", fd);

    /* use file descriptor, because fread hangs here? */
    if (TEMP_FAILURE_RETRY(recv(fd, &header, sizeof(read_header), 0)) < (signed) sizeof(read_header))
    {
        F_("Could not read header.\n");
        return;
    }

    if (header.p_ver != PROTOCOL_4_VERSION)
    {
        F_("ngc protcol_version miss-match, server_protocol_version :%i, client_protocol_version :%i !\n Will try to hot-reload initng.", PROTOCOL_4_VERSION, header.p_ver);
	initng_reload();
        return;
    }

    header.l[100] = '\0';
    D_("command type '%c', long \"%s\", protocol_version %i\n", header.c, header.l, header.p_ver);
    if(header.body_len > 0)
    {
	D_("There is a body to read!\n");
	header_data = i_calloc(1, header.body_len + 1);
	if(!header_data)
	{
	    F_("Coud not allocate memory fore header_data\n");
	    return;
	}
	
	/* yost so ngc got a chance to fill header_data again */
	usleep(10);
	
	if (TEMP_FAILURE_RETRY(recv(fd, header_data,  header.body_len, 0)) < (signed) header.body_len)
	{
    	    F_("Could not read header_data\n");
	    if(header_data) free(header_data);
    	    return;
	}
    }
	
    /* allocate space for the result we will send */
    result = i_calloc(1, sizeof(result_desc));
    if(!result)
	return;
	
    /* set version */
    strncpy(result->version, INITNG_VERSION, 100);
    printf("result->version: %s\n", result->version);
    result->p_ver = PROTOCOL_4_VERSION;

    /* ping check : */
    if (header.c == 'X')
    {
        result->c = 'Y';
        result->t = VOID_COMMAND;
        result->s = S_TRUE;
	result->payload = 0;
        D_("Ping received, sending pong\n");
	send(fd, result, sizeof(result_desc), 0);
	if(header_data) free(header_data);
	if(result) free(result);
        return;
    }

    /* Find the command requesting in the command database */
    if (header.c) /* find by short opt */
        tmp_cmd = initng_command_find_by_command_id(header.c);
    else if (header.l) /* find by long opt */
        tmp_cmd = initng_command_find_by_command_string(header.l);
    else
    {
	if(header_data) free(header_data);
	if(result) free(result);
        return;
    }

    /* Make sure the command we got is valid, else return an bad result */
    if (!tmp_cmd || tmp_cmd->com_type == 0)
    {
        D_("command type '%c', long \"%s\"\n", header.c,
           header.l);
        result->c = header.c;
        result->t = COMMAND_FAIL;
	result->payload = 0;
        result->s = S_COMMAND_NOT_FOUND;
        send(fd, result, sizeof(result_desc), 0);
	if(header_data) free(header_data);
	if(result) free(result);
        return;
    }

    /* check if command requires option, and option is not set */
    if (tmp_cmd->opt_type == REQUIRES_OPT && header.body_len < 1)
    {
        D_("Command %c - %s, requires an option!\n", header.c, header.l);
        result->c = header.c;
        result->t = COMMAND_FAIL;
	result->payload = 0;
        result->s = S_REQUIRES_OPT;
        send(fd, result, sizeof(result_desc), 0);
	if(header_data) free(header_data);
        return;
    }

    /* check if command is not using and option, and option is set */
    if (tmp_cmd->opt_type == NO_OPT && header.body_len > 0)
    {
        D_("Command %c - %s, don't want an option!\n", header.c, header.l);
        result->c = header.c;
        result->t = COMMAND_FAIL;
	result->payload = 0;
        result->s = S_NOT_REQUIRES_OPT;
        send(fd, result, sizeof(result_desc), 0);
	if(header_data) free(header_data);
        return;
    }

    /* set the result statics. */
    result->c = tmp_cmd->command_id;
    result->t = tmp_cmd->com_type;

    switch (tmp_cmd->com_type)
    {

        case TRUE_OR_FALSE_COMMAND:
        case INT_COMMAND:
            {
                int ret = 0;
	
                assert(tmp_cmd->u.int_command_call);
                D_("Calling an int or true or false command.\n");

                /* execute command */
                ret = (int) (*tmp_cmd->u.int_command_call) ((char *) header_data );
		
                /* Write a header respond */
                result->s = S_TRUE;
		result->payload = sizeof(int);
		
		/* send the result */
    		if ((sent=send(fd, result, sizeof(result_desc), 0)) != sizeof(result_desc))
		{
                    F_("failed to send result, sent: %i of %i.\n", sent, sizeof(result_desc));
		    break;
		}
		
		/* send the payload */
    		if ((sent=send(fd, &ret, result->payload, 0)) != (signed) result->payload)
		{
		    F_("Coud not send complete payload, sent %i of %i.", sent, result->payload);
		    break;
		}
                break;
            }
        case STRING_COMMAND:
            {
                char *send_buf = NULL;

                assert(tmp_cmd->u.string_command_call);
                D_("Calling an string command.\n");

                /* execute command */
                send_buf = (*tmp_cmd->u.string_command_call) ((char *) header_data);
		if(!send_buf)
		    break;
		    
		printf("Sending: %s\n", send_buf);
                /* write an header respond */
                result->s = S_TRUE;
		result->payload = strlen(send_buf) * sizeof(char);

		/* send the result */
    		if ((sent=send(fd, result, sizeof(result_desc), 0)) != sizeof(result_desc))
		{
                    F_("failed to send result, sent: %i of %i.\n", sent, sizeof(result_desc));
		    break;
		}
		/* send the payload */
    		if ((sent=send(fd, send_buf, result->payload, 0)) != (signed) result->payload)
		{
		    F_("Coud not send complete payload, sent %i of %i.", sent, result->payload);
		    break;
		}
		
		/* free and clear */
		free(send_buf);
                break;
            }
        case VOID_COMMAND:
            {
                assert(tmp_cmd->u.void_command_call);
                D_("Calling an void command!\n");

		/* execute command */
                (*tmp_cmd->u.void_command_call) ((char *) header_data);

                /* write an header respond */
                result->s = S_TRUE;
		result->payload=0; /* unknown payload size here, TODO FIX THIS */

		/* sent result */
    		if ((sent=send(fd, result, sizeof(result_desc), 0)) != sizeof(result_desc))
		{
                    F_("failed to send result, sent: %i of %i.\n", sent, sizeof(result_desc));
		    break;
		}
                break;
            }
        case PAYLOAD_COMMAND:
	case DATA_COMMANDS_ROW:
	case DATA_ACTIVE_ROW:
	case DATA_OPTION_ROW:
            {
		/* iniziate a data payload variable */
		s_payload payload;
		
		/* clear payload */
		memset(&payload, 0, sizeof(s_payload));
		
                assert(tmp_cmd->u.data_command_call);
                D_("Calling an data_command.\n");

                /* execute command */
                (*tmp_cmd->u.data_command_call) ((char *) header_data, &payload);


                /* write an header respond */
                result->s = S_TRUE;
		result->payload=payload.s;
		
		/* send the result */
    		if((sent=send(fd, result, sizeof(result_desc), 0)) != sizeof(result_desc))
		{
                    F_("failed to send result, sent: %i of %i.\n", sent, sizeof(result_desc));
		    break;
		}
		
		/* with out this, the client wont get the 2ond send. WHY? */
		usleep(1);
		
		/* send the payload */
		D_("Sending a payload of %i bytes.\n", result->payload);
    		if ((sent=send(fd, payload.p, result->payload, 0)) != (signed) result->payload)
		{
		    F_("Coud not send complete payload, sent %i of %i.", sent, result->payload);
		    break;
		}
		
		/* cleanup and free */
		free(payload.p);
                break;
            }
        case COMMAND_FAIL:

            /* return FAIL header respond */
            result->s = S_INVALID_TYPE;
    	    if ((sent=send(fd, result, sizeof(result_desc), 0)) != (signed) sizeof(result_desc))
	    {
                F_("failed to send result, sent: %i of %i.\n", sent, sizeof(result_desc));
		break;
	    }
            /* TODO: really continue?? */

            D_("Invalid command type '%c', line '%s'\n",
               header.c, header.l);
	    if(header_data) free(header_data);
            return;
    }
    D_("Returned successfully.\n");
    if(header_data) free(header_data);
    if(result) free(result);
    return;
}


/* called by fd hook, when data is no socket */
void accepted_client(f_module_h * from)
{
    int newsock;

    /* make a dumb check */
    if (from != &fdh)
        return;

    D_("Got here from fd hook.\n");
    /* we try to fix socket after every service start
       if it fails here chances are a user screwed it
       up, and we shouldn't manually try to fix anything. */
    if (fdh.fds <= 0)
    {
        /* socket is invalid but we were called, call closesocket to make sure it wont happen again */
        F_("accepted client called with fdh.fds %d, report bug\n", fdh.fds);
        closesock();
        F_("Attempting to reopen socket\n");
        open_socket();
        return;
    }

    /* create a new socket, for reading */
    if ((newsock = accept(fdh.fds, NULL, NULL)) > 0)
    {
        /*file_sock = fdopen(newsock, "r+");
        D_("read socket open, now reading\n");*/
	
        /* read the data, by the handle_client function */
        handle_client(newsock);
        
	/*close(newsock);*/
	/*  clean up, and return
	    fclose(file_sock); */
        D_("Success!\n");
        return;
    }

    /* temporary unavailable */
    if ((errno == EAGAIN) || (errno == EWOULDBLOCK))
    {
        W_("errno = EAGAIN!\n");
        return;
    }

    /* This'll generally happen on shutdown, don't cry about it. */
    D_("Error accepting socket %d, %s\n", fdh.fds, strerror(errno));
    closesock();
    return;
}

/* Send a ping to ourselves to check if we're 100% ok. */
static int sendping()
{
    int client;
    struct sockaddr_un sockname;
    read_header header;
    result_desc result;


    D_("Sending ping\n");

    memset(&header, 0, sizeof(read_header));
    header.p_ver = PROTOCOL_4_VERSION;

    /* Create the socket. */
    client = socket(PF_UNIX, SOCK_STREAM, 0);
    if (client < 0)
    {
        F_("Failed to init socket\n");
        return FALSE;
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;

    strcpy(sockname.sun_path, socket_filename);

    /* Try to connect */
    if (connect
        (client, (struct sockaddr *) &sockname,
         (strlen(sockname.sun_path) + sizeof(sockname.sun_family))) < 0)
    {
        close(client);
        return FALSE;
    }

    /* Write X to ping ourselves */

    header.c = 'X';
    header.l[0] = '\0';

    D_("Sending PING..\n");
    if (write(client, &header, sizeof(read_header)) <
        (signed) sizeof(read_header))
    {
        F_("Unable to send PING!\n");
        close(client);
        return (FALSE);
    }
    D_("PING sent..\n");

    /* Accept "server side" */
    accepted_client(&fdh);

    D_("Reading PONG..\n");
    if ((read(client, &result, sizeof(result_desc)) <
         (signed) sizeof(result_desc)) || result.c != 'Y'
        || result.s != S_TRUE)
    {
        F_("Unable to receive PONG!\n");
        close(client);
        return (FALSE);
    }

    D_("Got pong\n");

    return TRUE;
}

/* this will try to open a new socket */
static int open_socket()
{
    /*    int flags; */
    struct sockaddr_un serv_sockname;

    D_("Creating " SOCKET_4_ROOTPATH " dir\n");

    closesock();

    /* Make /dev/initng if it doesn't exist (try either way) */
    if (mkdir(SOCKET_4_ROOTPATH, S_IRUSR | S_IWUSR | S_IXUSR) == -1
        && errno != EEXIST)
    {
        if (errno != EROFS)
            F_("Could not create " SOCKET_4_ROOTPATH
               " : %s, may be / fs not mounted read-write yet?, will retry until I succeed.\n",
               strerror(errno));
        return (FALSE);
    }

    /* chmod root path for root use only */
    if (chmod(SOCKET_4_ROOTPATH, S_IRUSR | S_IWUSR | S_IXUSR) == -1)
    {
        /* path doesn't exist, we don't have /dev yet. */
        if (errno == ENOENT || errno == EROFS)
            return (FALSE);

        F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
           SOCKET_4_ROOTPATH);
    }

    /* Create the socket. */
    fdh.fds = socket(PF_UNIX, SOCK_STREAM, 0);
    if (fdh.fds < 1)
    {
        F_("Failed to init socket (%s)\n", strerror(errno));
        fdh.fds = -1;
        return (FALSE);
    }

    /* Set socket to non blocking mode */
    /*    flags = fcntl(fdh.fds, F_GETFL);
       if (flags < 0)
       {
       F_("Failed to fcntl fdh.fds\n");
       closesock();
       return (FALSE);
       }
       flags |= O_NONBLOCK;
       if (fcntl(fdh.fds, F_SETFL, flags) < 0)
       {
       F_("Failed to set fdh.fds O_NONBLOCK\n");
       closesock();
       return (FALSE);
       } */


    /* Bind a name to the socket. */
    serv_sockname.sun_family = AF_UNIX;

    strcpy(serv_sockname.sun_path, socket_filename);

    /* Remove old socket file if any */
    unlink(serv_sockname.sun_path);

    /* Try to bind */
    if (bind
        (fdh.fds, (struct sockaddr *) &serv_sockname,
         (strlen(serv_sockname.sun_path) +
          sizeof(serv_sockname.sun_family))) < 0)
    {
        F_("Error binding to socket (errno: %d str: '%s')\n", errno,
           strerror(errno));
        closesock();
        unlink(serv_sockname.sun_path);
        return (FALSE);
    }

    /* chmod socket for root only use */
    if (chmod(serv_sockname.sun_path, S_IRUSR | S_IWUSR) == -1)
    {
        F_("CRITICAL, failed to chmod %s, THIS IS A SECURITY PROBLEM.\n",
           serv_sockname.sun_path);
        closesock();
        return (FALSE);
    }

    /* store sock_stat for checking if we need to recreate socket later */
    stat(serv_sockname.sun_path, &sock_stat);

    /* Listen to socket */
    if (listen(fdh.fds, 5))
    {
        F_("Error on listen (errno: %d str: '%s')\n", errno, strerror(errno));
        closesock();
        unlink(serv_sockname.sun_path);
        return (FALSE);
    }



    /* Run check : */
    if (!sendping())
    {
        F_("Sendping check failed, ngc2 communication not available (if you see this open a bug)\n");
        closesock();
        return (FALSE);
    }

    return (TRUE);
}

/* this will check socket, and reopen on failure */
static void check_socket()
{
    struct stat st;

    D_("Checking socket\n");

    /* Check if socket needs reopening */
    if (fdh.fds <= 0)
    {
        D_("fdh.fds not set, opening new socket.\n");
        open_socket();
        return;
    }

    /* stat the socket, reopen on failure */
    memset(&st, 0, sizeof(st));
    if (stat(socket_filename, &st) < 0)
    {
        W_("Stat failed! Opening new socket.\n");
        open_socket();
        return;
    }

    /* compare socket file, with the one that we know, reopen on failure */
    if (st.st_dev != sock_stat.st_dev || st.st_ino != sock_stat.st_ino
        || st.st_mtime != sock_stat.st_mtime)
    {
        F_("Invalid socket found, reopening\n");
        open_socket();
        return;
    }

    D_("Socket ok.\n");
    return;
}


int module_init(int api_version)
{
    D_("module_init(ngc2);\n");
    if (api_version != API_VERSION)
    {
        F_("This module is compiled for api_version %i version and initng is compiled on %i version, won't load this module!\n", INITNG_VERSION, api_version);
        return (FALSE);
    }

    /* zero globals */
    fdh.fds = -1;
    memset(&sock_stat, 0, sizeof(sock_stat));


    /* decide which socket to use */
    /*if (g.i_am_init)*/
        socket_filename = SOCKET_4_FILENAME_REAL;
    /*else
        socket_filename = SOCKET_4_FILENAME_TEST;*/
    D_("Socket is: %s\n", socket_filename);

    D_("adding hook, that will reopen socket, for every started service.\n");
    initng_plugin_hook_add(&g.FDWATCHERS, 30, &fdh);
    initng_plugin_hook_add(&g.HUP, 50, &check_socket);

    /* do the first socket directly */
    open_socket();

    D_("ngc2.so.0.0 loaded!\n");
    return (TRUE);
}


void module_unload(void)
{
    D_("module_unload(ngc2);\n");

    /* close open sockets */
    closesock();

    /* remove hooks */
    initng_plugin_hook_del(&g.FDWATCHERS, &fdh);
    initng_plugin_hook_del(&g.HUP, &check_socket);

    D_("ngc2.so.0.0 unloaded!\n");

}
